import os
import sys
import platform
import importlib
import traceback
import importlib.util
import importlib.machinery
from pathlib import Path
from contextlib import contextmanager

from kubiya_sdk.tools.models import Tool
from kubiya_sdk.tools.registry import tool_registry
from kubiya_sdk.utils.logger import sdk_logger

from .models import BundleModel
from rich.console import Console
from rich.table import Table
from rich.text import Text


class CustomModuleLoader:
    def __init__(self, directory):
        self.directory = directory

    def load_module(self, fullname):
        if fullname in sys.modules:
            return sys.modules[fullname]

        parts = fullname.split(".")
        path = self.directory
        for part in parts:
            path = os.path.join(path, part)

        if os.path.isdir(path):
            path = os.path.join(path, "__init__.py")
        else:
            path += ".py"

        if not os.path.exists(path):
            raise ImportError(f"No module named '{fullname}'")

        spec = importlib.util.spec_from_file_location(fullname, path)
        module = importlib.util.module_from_spec(spec)
        sys.modules[fullname] = module
        spec.loader.exec_module(module)

        return module


class CustomPathFinder:
    def __init__(self, directory):
        self.directory = directory
        self.loader = CustomModuleLoader(directory)

    def find_spec(self, fullname, path, target=None):
        parts = fullname.split(".")
        path = self.directory
        for part in parts:
            path = os.path.join(path, part)

        if os.path.isdir(path) or os.path.exists(path + ".py"):
            return importlib.machinery.ModuleSpec(fullname, self.loader)

        return None


@contextmanager
def add_custom_importer(directory):
    custom_finder = CustomPathFinder(directory)
    sys.meta_path.insert(0, custom_finder)
    try:
        yield
    finally:
        sys.meta_path.remove(custom_finder)


def import_module_from_path(module_path, module_name):
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def import_all_modules(directory, ignore_dirs: tuple[str] = tuple()) -> list[str]:
    errors = []
    with add_custom_importer(directory):
        for root, dirs, files in os.walk(directory):
            directories_to_ignore = ignore_dirs + (".venv", "venv")
            for dir_to_remove in directories_to_ignore:
                if dir_to_remove in dirs:
                    sdk_logger.log(
                        f"Removing {dir_to_remove} from dirs",
                        component="discovery",
                        level="INFO",
                    )
                    dirs.remove(dir_to_remove)
            for file in files:
                if file.endswith(".py") and not file.startswith("__"):
                    module_path = os.path.join(root, file)
                    sdk_logger.log(
                        f"Importing {module_path}",
                        component="discovery",
                        level="DEBUG",
                    )
                    module_name = os.path.relpath(
                        os.path.splitext(module_path)[0], directory
                    ).replace(os.path.sep, ".")
                    try:
                        import_module_from_path(module_path, module_name)
                        sdk_logger.log(
                            f"Imported {module_path}",
                            component="discovery",
                            level="DEBUG",
                        )
                    except Exception as e:
                        error = f"Failed to import {module_path}: {e}"
                        sdk_logger.log(
                            error,
                            component="discovery",
                            level="WARNING",
                        )
                        errors.append(error)

    return errors


def get_tools(
    base_dir: str, ignore_dirs: tuple[str] = tuple()
) -> tuple[list[Tool], list[str]]:
    sdk_logger.log(
        f"Discovering workflows and tools in directory: {base_dir}",
        component="discovery",
    )
    errors = []
    try:
        import_errors = import_all_modules(base_dir, ignore_dirs)
    except BaseException as e:
        traceback.print_exc()
        sdk_logger.log(
            f"Exception while importing modules",
            component="Bundle",
            level="ERROR",
        )
        raise RuntimeError("Failed to import modules") from e

    errors.extend(import_errors)
    tools = []
    for _, tools_dict in tool_registry.tools.items():
        for tool_name, tool in tools_dict.items():
            try:
                if not tool.mermaid:
                    sdk_logger.log(
                        f"Tool {tool_name} doesn't have a mermaid diagram. Generating one...",
                        component="discovery",
                        level="WARNING",
                    )
                    tool.mermaid = tool.to_mermaid()
                tools.append(tool)
            except Exception as tool_error:
                msg = f"Error processing tool {tool_name}: {str(tool_error)}"
                sdk_logger.log(
                    msg,
                    component="discovery",
                    level="ERROR",
                )
                errors.append(msg)
    return tools, errors


def bundle(base_dir: str, ignore_dirs: tuple[str] = tuple()) -> dict:
    tools, errors = get_tools(base_dir, ignore_dirs)

    result = {
        "python_version": platform.python_version(),
        # "package_name": os.path.basename(source),
        # "package_root_path": source,
        "workflows": [],
        "tools": [tool.model_dump(exclude_unset=True) for tool in tools],
        "errors": errors,
        # "source": source_info,
    }
    bundle = BundleModel(
        tools=tools,
        errors=errors,
        python_bundle_version=platform.python_version(),
    )
    bundle_file = Path(base_dir) / "kubiya_bundle.json"
    try:
        bundle_file.write_text(bundle.model_dump_json(indent=2))
    except Exception as e:
        sdk_logger.log(
            f"Failed to write bundle file to {bundle_file}: {str(e)}",
            component="discovery",
            level="ERROR",
        )

    console = Console()

    # Print a pretty bundle summary
    console.rule("[bold blue]Bundle Summary[/bold blue]")

    console.print(
        f"[bold]Python Version:[/bold] [bold green]{result['python_version']}[/bold green]"
    )
    console.print(
        f"[bold]Tools Discovered:[/bold] [bold green]{len(tools)}[/bold green]"
    )

    table = Table(title="Tools Summary", show_header=True, header_style="bold magenta")
    table.add_column("Tool Name", style="bold", width=30)
    table.add_column("Args", justify="center")
    table.add_column("Env Vars", justify="center")
    table.add_column("Secrets", justify="center")
    table.add_column("Files", justify="center")

    for tool in tools:
        table.add_row(
            tool.name,
            str(len(tool.args)),
            str(len(tool.env)),
            str(len(tool.secrets) if tool.secrets is not None else 0),
            str(len(tool.with_files) if tool.with_files is not None else 0),
        )

    console.print(table)

    if errors:
        console.print(f"[bold red]Number of Errors: {len(errors)}[/bold red]")
        for error in errors:
            console.print(f"[red]Error:[/red] {error}")
    else:
        console.print("[bold green]No Errors[/bold green]")

    console.rule("[bold blue]End of Summary[/bold blue]")
    return result
